@@ -5,7 +5,7 @@
 # Copyright (C) 2002       Masato Taruishi <taru@debian.org>
 # Copyright (C) 2006-2008  Junichi Uekawa <dancer@debian.org>
 # Copyright (C) 2007       Famelis George <famelis@otenet.gr>
-# Copyright (C) 2008-2019  Francesco Poli <invernomuto@paranoici.org>
+# Copyright (C) 2008-2020  Francesco Poli <invernomuto@paranoici.org>
 # Copyright (C) 2009-2010  Ryan Niebur <ryan@debian.org>
 # Copyright (C) 2013       Google Inc
 #
@@ -44,10 +44,9 @@ class AppConfig
   SENSIBLE_BROWSER = "/usr/bin/sensible-browser"
   XDG_OPEN = "/usr/bin/xdg-open"
   WWW_BROWSER = "/usr/bin/www-browser"
-  S6_SETUIDGID = "/usr/bin/s6-setuidgid"
 
-  def usage
-    $stderr.print _("Usage: "), File.basename($0),
+  def usage(stream = $stderr)
+    stream.print _("Usage: "), File.basename($0),
       _(" [options] <command> [arguments]"),
       "\n",
       _("Options:\n"),
@@ -184,7 +183,7 @@ class AppConfig
       opt_parser.each_option { |optname, optargs|
         case optname
         when '--help'
-          usage
+          usage($stdout)
           exit 0
         when '--version'
           puts $VERSION
@@ -267,9 +266,9 @@ class AppConfig
 
     # proxy settings in apt.conf
     if /http_proxy='(.*)'/ =~ `apt-config #{@apt_conf} shell http_proxy acquire::http::proxy`
-      puts "proxy configuration from apt.conf: #{$1}" if $DEBUG
+      $stderr.puts "proxy configuration from apt.conf: #{$1}" if $DEBUG
       if $1 == 'DIRECT' || $1 == ''
-        puts "Disabling proxy due to DIRECT, or empty string" if $DEBUG
+        $stderr.puts "Disabling proxy due to DIRECT, or empty string" if $DEBUG
         ENV.delete("http_proxy")
         ENV.delete("soap_use_proxy")
       else
@@ -278,24 +277,24 @@ class AppConfig
       end
     end
     if /proxy_detect='(.*)'/ =~ `apt-config #{@apt_conf} shell proxy_detect acquire::http::proxy-auto-detect`
-      puts "auto proxy detect command from apt.conf: #{$1}" if $DEBUG
+      $stderr.puts "auto proxy detect command from apt.conf: #{$1}" if $DEBUG
       if FileTest.executable?("#{$1}")
         detected_proxy = `#{$1}`.chomp
-        puts "auto detected proxy: #{detected_proxy}" if $DEBUG
+        $stderr.puts "auto detected proxy: #{detected_proxy}" if $DEBUG
         if /^http/ =~ detected_proxy
           ENV["http_proxy"] = detected_proxy
           ENV["soap_use_proxy"] = "on"
         else
-          puts "Ignoring auto detected proxy not beginning with 'http'" if $DEBUG
+          $stderr.puts "Ignoring auto detected proxy not beginning with 'http'" if $DEBUG
         end
       else
-        STDERR.puts _("W: ") + sprintf(_("Cannot execute auto proxy detect command %s"), $1)
+        $stderr.puts _("W: ") + sprintf(_("Cannot execute auto proxy detect command %s"), $1)
       end
     end
     if /http_proxy='(.*)'/ =~ `apt-config #{@apt_conf} shell http_proxy acquire::http::proxy::bugs.debian.org`
-      puts "proxy configuration from apt.conf, specific for bugs.debian.org: #{$1}" if $DEBUG
+      $stderr.puts "proxy configuration from apt.conf, specific for bugs.debian.org: #{$1}" if $DEBUG
       if $1 == 'DIRECT'
-        puts "Disabling proxy due to DIRECT" if $DEBUG
+        $stderr.puts "Disabling proxy due to DIRECT" if $DEBUG
         ENV.delete("http_proxy")
         ENV.delete("soap_use_proxy")
       else
@@ -308,7 +307,7 @@ class AppConfig
     command = ARGV.shift
     case command
     when nil
-      STDERR.puts _("E: ") + _("You need to specify a command.")
+      $stderr.puts _("E: ") + _("You need to specify a command.")
       usage
       exit 1
     when "list"
@@ -318,7 +317,7 @@ class AppConfig
     when "rss"
       @command = "rss"
     else
-      STDERR.puts _("E: ") + _("Unknown command ") +  "'#{command}'."
+      $stderr.puts _("E: ") + _("Unknown command ") +  "'#{command}'."
       usage
       exit 1
     end
@@ -338,24 +337,15 @@ class AppConfig
       @browser = WWW_BROWSER
     end
 
-    setuidgid = nil
-    if FileTest.executable?("#{S6_SETUIDGID}")
-      setuidgid = S6_SETUIDGID
-    end
-
     if Process.euid == 0
       # running as root
       # external programs are better run after dropping root privileges
+      # and without accessing the graphical environment
       if @ext_user != "root"
         # running as root, after becoming root (with "su -", "sudo", ...)
         # the original user name is @ext_user
-        if setuidgid != nil
-          # a command to drop root privileges is available
-          @cmd_prefix = "#{setuidgid} #{@ext_user} "
-        else
-          # no command to drop root privileges found
-          @ext_user = "root"
-        end
+        # use setpriv to drop root privileges
+        @cmd_prefix = "env -u DISPLAY -u XAUTHORITY setpriv --reuid #{@ext_user} --init-groups "
       end
     else
       # not running as root
@@ -448,12 +438,13 @@ class Viewer
 
   class SimpleViewer < Viewer
 
-    DeprecatedWarning = _("********** on_hold IS DEPRECATED. USE p INSTEAD to use pin **********")
-    DeprecatedWarningHeader = "*" * Unicode.width(DeprecatedWarning)
-
     def initialize(config)
       super(config)
       @bugmap = {}
+      @hold_pkg_keys = []
+      @dodged_bugnums = []
+      @ignored_bugnums = []
+      @pref_str = ""
     end
 
     def view(new_pkgs, cur_pkgs, bugs)
@@ -466,10 +457,9 @@ class Viewer
       end
 
       answer = "n"
-      hold_pkg_keys = []
       while true
         ask_str = _("Are you sure you want to install/upgrade the above packages?").dup
-        if hold_pkg_keys.empty?
+        if @hold_pkg_keys.empty? && @dodged_bugnums.empty?
           ask_str << " [Y/n/?/...]"
         else
           ask_str << " [N/?/...]"
@@ -493,7 +483,7 @@ class Viewer
           @config.frontend.puts ""
         end
         if a == ""
-          if hold_pkg_keys.empty?
+          if @hold_pkg_keys.empty? && @dodged_bugnums.empty?
             answer = "y"
           else
             answer = "n"
@@ -504,20 +494,13 @@ class Viewer
         case answer
 
         when "y"
-          return true if hold_pkg_keys.empty?
-
-        when "a"
-          if hold_pkg_keys.empty?
-            bugs.each { |bug|
-              if ! @config.system_ignore_bugs.include?(bug.bug_number)
-                @config.system_ignore_bugs.add(bug)
-                @config.system_ignore_bugs.add(bug.bug_number)
-              end
-            }
+          if @hold_pkg_keys.empty? && @dodged_bugnums.empty?
+            save(bugs)
             return true
           end
 
         when "n"
+          save(bugs)
           return false
 
         when /^(#|b)?(\d+)$/
@@ -533,10 +516,10 @@ class Viewer
               bugnum = $2
             end
             if ! bugnum.nil?
-              puts "Invoking querybts for ##{bugnum}" if $DEBUG
+              $stderr.puts "Invoking querybts for ##{bugnum}" if $DEBUG
               querybtscommandline = "#{@config.cmd_prefix}#{@config.querybts} -u text #{bugnum}"
               if system(querybtscommandline)
-                puts "successfully invoked querybts" if $DEBUG
+                $stderr.puts "successfully invoked querybts" if $DEBUG
               else
                 $stderr.puts _("W: ") + _("Failed to invoke querybts.")
                 $stderr.puts " #{querybtscommandline}"
@@ -558,99 +541,148 @@ class Viewer
             bugnum = $2
           end
           if ! bugnum.nil?
-            if ! @config.system_ignore_bugs.include?(bugnum)
-              bug = @config.parser.parse_bug(bugnum)
-              @config.system_ignore_bugs.add(bug)
-              @config.system_ignore_bugs.add(bugnum)
-              Factory::BugsFactory.delete_ignore_bugs(bugs)
-              # TRANSLATORS: "ignored" refers to one singular bug
-              @config.frontend.puts sprintf(_("%s ignored"), bugnum)
-            else
-              # TRANSLATORS: "ignored" refers to one singular bug
-              @config.frontend.puts sprintf(_("%s already ignored"), bugnum)
-            end
+            bug = bugs.extract(bugnum)
+            bug = @config.parser.parse_bug(bugnum) if bug.nil?
+            bugs_to_ignore = [bugnum]
+            bugs_to_ignore.concat(bug.mergeids)
+            bugs_to_ignore.each { |bugnum|
+              if ! @config.system_ignore_bugs.include?(bugnum) && ! @ignored_bugnums.include?(bugnum)
+                @ignored_bugnums << bugnum
+                # TRANSLATORS: "ignored" refers to one singular bug
+                @config.frontend.puts sprintf(_("%s ignored"), bugnum)
+              else
+                # TRANSLATORS: "ignored" refers to one singular bug
+                @config.frontend.puts sprintf(_("%s already ignored"), bugnum)
+              end
+            }
           end
 
+        when "i"
+          bugs.filter_out(@dodged_bugnums+@ignored_bugnums).each { |bug|
+            if ! @hold_pkg_keys.include?(bug.pkg_key)
+              if ! @config.system_ignore_bugs.include?(bug.bug_number) && ! @ignored_bugnums.include?(bug.bug_number)
+                @ignored_bugnums << bug.bug_number
+                @config.frontend.puts sprintf(_("%s ignored"), bug.bug_number)
+              end
+            end
+          }
+
         when "r"
-          display_bugs(bugs, new_pkgs.keys - hold_pkg_keys, cur_pkgs, new_pkgs)
+          display_bugs(bugs, new_pkgs.keys, cur_pkgs, new_pkgs)
 
-        when /^(h|p)\s+(.+)$/
-          cmd = $1
-          if cmd == "h"
-            @config.frontend.puts DeprecatedWarningHeader
-            @config.frontend.puts DeprecatedWarning
-            @config.frontend.puts DeprecatedWarningHeader
+        when /^p\s+(.+)$/
+          pkg_keys = $1.split(/\s+/)
+          # TRANSLATORS: %{plist} is a comma-separated list of %{npkgs} packages to be pinned.
+          if @config.frontend.yes_or_no? ngettext(
+           "The following %{npkgs} package will be pinned:\n %{plist}\nAre you sure?",
+           "The following %{npkgs} packages will be pinned:\n %{plist}\nAre you sure?",
+           pkg_keys.size) % {:npkgs => pkg_keys.size,
+                             :plist => pkg_keys.join(', ')}
+            h, e = pinned(pkg_keys, cur_pkgs, bugs)
+            @hold_pkg_keys.concat(h) if h != nil
           end
-          pkg_keys = $2.split(/\s+/)
-          if cmd == "h"
-            h = on_hold(pkg_keys)
-          else
-            h = pinned(pkg_keys, cur_pkgs, bugs)
+
+        when /^d\s+(.+)$/
+          bug_names = $1.split(/\s+/)
+          bugnames = []
+          bugnums  = []
+          bug_names.each { |name|
+            bugnum = nil
+            if /^(b)?(\d+)$/ =~ name
+              if $1 == "b"
+                bugnum = @bugmap[$2] if @bugmap.has_key?($2)
+              else
+                bugnum =         $2  if @bugmap.has_value?($2)
+              end
+            end
+            if bugnum.nil?
+              @config.frontend.puts sprintf(_("%s is unknown"), "'#{name}'")
+            else
+              bugnames << name
+              bugnums  << bugnum
+              bug = bugs.extract(bugnum)
+              bugnums.concat(bug.mergeids) unless bug.nil?
+            end
+          }
+          feared_bugs = bugs.filter_out(@dodged_bugnums+@ignored_bugnums).filter_in(bugnums)
+          pkgs = {}
+          feared_bugs.each { |bug|
+            pkgs[bug.pkg_key] = 1 unless @hold_pkg_keys.include?(bug.pkg_key)
+          }
+          if pkgs.size != 0
+            # TRANSLATORS: %{blist} is a comma-separated list of %{nbugs} bugs to be dodged.
+            if @config.frontend.yes_or_no? ngettext(
+             "The following %{nbugs} bug will be dodged:\n %{blist}\nAre you sure?",
+             "The following %{nbugs} bugs will be dodged:\n %{blist}\nAre you sure?",
+             bugnames.size) % {:nbugs => bugnames.size,
+                               :blist => bugnames.join(', ')}
+              h, e = pinned(pkgs.keys, cur_pkgs, feared_bugs)
+            end
           end
-          hold_pkg_keys.concat(h) if h != nil
+          @dodged_bugnums.concat(e) if e != nil
+
+        when "c"
+          $stderr.puts bugs if $DEBUG
+          display_bugs_as_html(bugs, cur_pkgs, new_pkgs, "c")
 
         when "w"
-          puts bugs if $DEBUG
+          $stderr.puts bugs if $DEBUG
           if @config.browser != nil
             display_bugs_as_html(bugs, cur_pkgs, new_pkgs)
           else
             @config.frontend.puts sprintf(_("You must install a web browser package to be able to do this"))
           end
 
-        when /^(h|p)$/
-          cmd = $1
-          if cmd == "h"
-            @config.frontend.puts DeprecatedWarningHeader
-            @config.frontend.puts DeprecatedWarning
-            @config.frontend.puts DeprecatedWarningHeader
-          end
+        when "p"
           pkgs = {}
-          if cmd == "p"
-            prev_pinned_pkgs = {}
-            begin
-              p = Debian::AptPreferences.new
-              p.pins.each { |pin|
-                prev_pinned_pkgs[pin["Package"]] = 1
-              }
-            rescue
-              # something went wrong while parsing preferences: assume no pins
-            end
-            bugs.each { |bug|
-              if ! prev_pinned_pkgs.has_key?(bug.pkg_key)
-                pkgs[bug.pkg_key] = 1
-              end
-            }
-          else
-            bugs.each { |bug|
-              pkgs[bug.pkg_key] = 1
+          prev_pinned_pkgs = {}
+          begin
+            p = Debian::AptPreferences.new
+            p.pins.each { |pin|
+              prev_pinned_pkgs[pin["Package"]] = 1
             }
+          rescue
+            # something went wrong while parsing preferences: assume no pins
           end
+          filtered_bugs = bugs.filter_out(@dodged_bugnums+@ignored_bugnums)
+          filtered_bugs.each { |bug|
+            if ! prev_pinned_pkgs.has_key?(bug.pkg_key)
+              pkgs[bug.pkg_key] = 1 unless @hold_pkg_keys.include?(bug.pkg_key)
+            end
+          }
           if pkgs.size != 0
             @config.yes = true if @config.force_pin
             if @config.frontend.yes_or_no? ngettext(
-             # TRANSLATORS: %{plist} is a comma-separated list of %{npkgs} packages to be pinned or put on hold.
-             "The following %{npkgs} package will be pinned or on hold:\n %{plist}\nAre you sure?",
-             "The following %{npkgs} packages will be pinned or on hold:\n %{plist}\nAre you sure?",
+             "The following %{npkgs} package will be pinned:\n %{plist}\nAre you sure?",
+             "The following %{npkgs} packages will be pinned:\n %{plist}\nAre you sure?",
              pkgs.size) % {:npkgs => pkgs.size,
                            :plist => pkgs.keys.join(', ')}
-              if cmd == "h"
-                h = on_hold(pkgs.keys)
-              else
-                h = pinned(pkgs.keys, cur_pkgs, bugs)
-              end
+              h, e = pinned(pkgs.keys, cur_pkgs, bugs)
             end
-            hold_pkg_keys.concat(h) if h != nil
+            @hold_pkg_keys.concat(h) if h != nil
           else
-            @config.frontend.puts sprintf(_("All selected packages are already pinned or on hold. Ignoring %s command."), cmd)
+            @config.frontend.puts sprintf(_("All selected packages are already pinned. Ignoring %s command."), answer)
           end
           return false if @config.force_pin
 
+        when "u"
+          if @hold_pkg_keys.empty? && @dodged_bugnums.empty? && @ignored_bugnums.empty?
+            @config.frontend.puts sprintf(_("There are no dodge/pin/ignore operations to undo. Ignoring %s command."), answer)
+          else
+            if @config.frontend.yes_or_no?(_("All the dodge/pin/ignore operations will be undone.\nAre you sure?"))
+              @hold_pkg_keys = []
+              @dodged_bugnums = []
+              @ignored_bugnums = []
+              @pref_str = ""
+              display_bugs(bugs, new_pkgs.keys, cur_pkgs, new_pkgs)
+            end
+          end
+
         else
-          if hold_pkg_keys.empty?
+          if @hold_pkg_keys.empty? && @dodged_bugnums.empty?
             @config.frontend.puts "" +
               # TRANSLATORS: the dashes (-) in the following strings are vertically aligned, please keep their alignment consistent
-              _("     y     - continue the APT installation, but do not mark the bugs\n             as ignored.\n") +
-              _("     a     - continue the APT installation and mark all the above bugs\n             as ignored.\n")
+              _("     y     - continue the APT installation.\n")
           end
           @config.frontend.puts "" +
             _("     n     - stop the APT installation.\n")
@@ -665,20 +697,45 @@ class Viewer
           end
           @config.frontend.puts "" +
             _("     r     - redisplay bug lists.\n") +
-            _(" p <pkg..> - pin pkgs (restart APT session to enable).\n") +
-            _(" p         - pin all the above pkgs (restart APT session to enable).\n") +
-            _(" i <num>   - mark bug number <num> as ignored.\n") +
-            _(" i b<id>   - mark the bug identified by <id> as ignored.\n") +
-            _("     ?     - print this help.\n")
+            _("     c     - compose bug lists in HTML.\n")
           if @config.browser != nil
             @config.frontend.puts _("     w     - display bug lists in HTML\n             (uses %{prog} as user %{user}).\n") %
               {:prog => File.basename(@config.browser),
                :user => @config.ext_user}
           end
+          @config.frontend.puts "" +
+            _(" d <num>.. - dodge bugs <num> by pinning affected pkgs\n             (restart APT session to enable).\n") +
+            _(" d b<id>.. - dodge bugs identified by <id> by pinning affected pkgs\n             (restart APT session to enable).\n") +
+            _(" p <pkg>.. - pin pkgs\n             (restart APT session to enable).\n") +
+            _(" p         - pin all the above pkgs\n             (restart APT session to enable).\n") +
+            _(" i <num>   - mark bug number <num> as ignored.\n") +
+            _(" i b<id>   - mark the bug identified by <id> as ignored.\n") +
+            _(" i         - mark all the above bugs as ignored.\n") +
+            _("     u     - undo all the dodge/pin/ignore operations done so far.\n") +
+            _("     ?     - print this help.\n")
         end
       end
     end
 
+    def save(bugs)
+      if @pref_str != ""
+        filename = "/etc/apt/preferences.d/apt-listbugs"
+        begin
+          File.open(filename, "a") { |io| io.puts @pref_str }
+        rescue Errno::EACCES
+          # write-access is not possible, warn the user that the
+          # file won't be updated
+          $stderr.puts _("W: ") + sprintf(_("Cannot write to %s"), filename)
+        end
+      end
+      @ignored_bugnums.each { |bugnum|
+        bug = bugs.extract(bugnum)
+        bug = @config.parser.parse_bug(bugnum) if bug.nil?
+        @config.system_ignore_bugs.add(bug)
+        @config.system_ignore_bugs.add(bugnum)
+      }
+    end
+
     def bugs_of_pkg( bugs, pkg_key )
       b = []
       bugs.each { |bug|
@@ -690,8 +747,11 @@ class Viewer
     def pinned(pkg_keys, cur_pkgs, bugs)
       holdstr = ""
       pinned_pkg_keys = []
-      pkg_keys.each { |pkg_key|
-        my_bugs = bugs_of_pkg( bugs, pkg_key )
+      explanation_bugs = []
+      notheld_pkg_keys = pkg_keys - @hold_pkg_keys
+      notheld_pkg_keys.each { |pkg_key|
+        my_bugs = bugs_of_pkg(bugs.filter_out(@dodged_bugnums+@ignored_bugnums),
+                              pkg_key)
         if my_bugs.size > 0 || @config.frontend.yes_or_no?(sprintf(
          _("None of the above bugs is assigned to package %s\nAre you sure you want to pin it?"),
          pkg_key))
@@ -707,6 +767,7 @@ class Viewer
           holdstr << "\nExplanation: Pinned by apt-listbugs at #{Time.now}"
           my_bugs.each { |bug|
             holdstr << "\nExplanation:   ##{bug.bug_number}: #{bug.desc}"
+            explanation_bugs << bug.bug_number
           }
           holdstr << "\nPackage: #{pkg_key}\nPin: version #{pin_ver}"
           holdstr << "\nPin-Priority: #{pin_pri}\n"
@@ -714,33 +775,11 @@ class Viewer
       }
       $stderr.puts holdstr if $DEBUG
       if holdstr != ""
-        filename = "/etc/apt/preferences.d/apt-listbugs"
-        begin
-          File.open(filename, "a") { |io|
-            io.puts holdstr
-            # TRANSLATORS: %{packgl} is a list of packages, %{filenm} is a file name.
-            @config.frontend.puts _("%{packgl} pinned by adding Pin preferences in %{filenm}. Restart APT session to enable") %
-              {:packgl => pinned_pkg_keys.join(' '),
-               :filenm => filename}
-            return pinned_pkg_keys
-          }
-        rescue Errno::EACCES
-          # write-access is not possible, warn the user that the
-          # file won't be updated
-          $stderr.puts _("W: ") + sprintf(_("Cannot write to %s"), filename)
-        end
-      end
-      return nil
-    end
-
-    def on_hold (pkg_keys)
-      holdstr = ""
-      pkg_keys.each { |pkg_key|
-        holdstr << "#{pkg_key} hold\n"
-      }
-      if system("echo '#{holdstr}' | dpkg --set-selections")
-        @config.frontend.puts sprintf(_("%s held. Restart APT session to enable"), pkg_keys.join(' '))
-        return pkg_keys
+        @pref_str << holdstr
+        # TRANSLATORS: %{packgl} is a list of packages.
+        @config.frontend.puts _("%{packgl} will be pinned. Restart APT session to enable") %
+              {:packgl => pinned_pkg_keys.join(' ')}
+        return pinned_pkg_keys, explanation_bugs
       end
       return nil
     end
@@ -748,14 +787,16 @@ class Viewer
     def display_bugs(bugs, pkg_keys, cur_pkgs, new_pkgs)
       # routine to display every bug that is available and relevant
 
+      filtered_bugs = bugs.filter_out(@dodged_bugnums+@ignored_bugnums)
+      notheld_pkg_keys = pkg_keys - @hold_pkg_keys
       bugs_statistics = {}
       @config.stats.each { |stat|
         @config.severity.each { |severity|
-          pkg_keys.each { |pkg_key|
+          notheld_pkg_keys.each { |pkg_key|
             p_bug_numbers = []
             bug_exist = 0
             bugs_statistics[pkg_key] = 0 unless bugs_statistics[pkg_key]
-            bugs.each_by_category(pkg_key, severity, stat) { |bug|
+            filtered_bugs.each_by_category(pkg_key, severity, stat) { |bug|
               next if p_bug_numbers.include?(bug.bug_number)
               bugs_statistics[pkg_key] += 1
               p_bug_numbers << bug.bug_number
@@ -832,7 +873,7 @@ class Viewer
       }
     end
 
-    def display_bugs_as_html(bugs, cur_pkgs, new_pkgs)
+    def display_bugs_as_html(bugs, cur_pkgs, new_pkgs, cmd = "w")
       tmp = Tempfile.new(["apt-listbugs", ".html"])
       tmp.chmod(0644)
       tmp.puts "<?xml version=\"1.0\" encoding=\"#{Locale.charset}\"?>"
@@ -850,7 +891,9 @@ class Viewer
       tmp.puts "  </div>\n\n  <div id=\"page_body\">"
       tmp.puts "    <hr class=\"nocss\" />"
 
-      each_state_table(tmp, bugs, @config.stats) { |bugs|
+      filtered_bugs = bugs.filter_out(@dodged_bugnums+@ignored_bugnums)
+      filtered_bugs.reject! { |bug| @hold_pkg_keys.include?(bug.pkg_key) }
+      each_state_table(tmp, filtered_bugs, @config.stats) { |bugs|
         p_bug_numbers = []
         bugs.each { |bug|
           next if p_bug_numbers.include?(bug.bug_number)
@@ -871,15 +914,22 @@ class Viewer
       tmp.puts "  </div>\n</body>\n</html>"
       tmp.close
 
-      puts "Invoking browser for #{tmp.path}" if $DEBUG
-      browsercommandline = "#{@config.cmd_prefix}#{@config.browser} #{tmp.path}"
-      if system(browsercommandline)
-        puts "successfully invoked browser" if $DEBUG
-      else
-        $stderr.puts _("W: ") + _("Failed to invoke browser.")
-        $stderr.puts " #{browsercommandline}"
+      case cmd
+      when "c"
+        @config.frontend.puts "" +
+          _("You can view the bug lists in HTML at the following URI:\n") +
+          "file://" + tmp.path + "\n"
+      when "w"
+        $stderr.puts "Invoking browser for #{tmp.path}" if $DEBUG
+        browsercommandline = "#{@config.cmd_prefix}#{@config.browser} #{tmp.path}"
+        if system(browsercommandline)
+          $stderr.puts "successfully invoked browser" if $DEBUG
+        else
+          $stderr.puts _("W: ") + _("Failed to invoke browser.")
+          $stderr.puts " #{browsercommandline}"
+        end
+        clear_stdin
       end
-      clear_stdin
     end
 
     private
@@ -1002,15 +1052,15 @@ module Factory
       # TRANSLATORS: this sentence, followed by the translation of "Done" (see above) should fit in less than 79 columns to work well with default width terminals
       yield _("Retrieving bug reports..."), "0%"
       begin
-        config.frontend.puts " Using QueryStep = #{config.querystep}" if $DEBUG
-        config.frontend.puts " Using ParseStep = #{config.parsestep}" if $DEBUG
+        $stderr.puts " Using QueryStep = #{config.querystep}" if $DEBUG
+        $stderr.puts " Using ParseStep = #{config.parsestep}" if $DEBUG
         # obtain the array of bugs
         bugs = config.parser.parse(ma_copies, config.querystep, config.parsestep, config.severity) { |pct|
           yield _("Retrieving bug reports..."), pct
         }
       rescue SOAP::HTTPStreamError => exception
         config.frontend.puts _(" Fail")
-        config.frontend.puts " Exception: " + exception.class.to_s if $DEBUG
+        $stderr.puts " Exception: " + exception.class.to_s if $DEBUG
         $stderr.puts _("E: ") + _("HTTP GET failed")
         retrycount -= 1
         if config.frontend.yes_or_no?(_("Retry downloading bug information?")) && retrycount > 0
@@ -1022,7 +1072,7 @@ module Factory
         bugs = []
       rescue SOAP::EmptyResponseError => exception
         config.frontend.puts _(" Fail")
-        config.frontend.puts " Exception: " + exception.class.to_s if $DEBUG
+        $stderr.puts " Exception: " + exception.class.to_s if $DEBUG
         $stderr.puts _("E: ") + _("Empty stream from SOAP")
         retrycount -= 1
         if config.frontend.yes_or_no?(_("Retry downloading bug information?")) && retrycount > 0
@@ -1034,9 +1084,9 @@ module Factory
         bugs = []
       rescue Exception => exception
         config.frontend.puts _(" Fail")
-        config.frontend.puts " Exception: " + exception.class.to_s if $DEBUG
+        $stderr.puts " Exception: " + exception.class.to_s if $DEBUG
         config.frontend.puts _("Error retrieving bug reports from the server with the following error message:")
-        config.frontend.puts _("E: ") + "#{$!}"
+        $stderr.puts _("E: ") + "#{$!}"
         if exception.kind_of? SocketError
           config.frontend.puts _("It appears that your network connection is down. Check network configuration and try again")
         else
@@ -1061,17 +1111,17 @@ module Factory
 
     def delete_ignore_bugs(bugs)
       # ignoring ignore_bugs
-      bugs.delete_if { |bug| config.system_ignore_bugs.include?(bug.bug_number)}
+      bugs.delete_if { |bug| config.system_ignore_bugs.include?(bug.bug_number) }
     end
 
     def delete_regexp_bugs(bugs, regexp)
-      puts "Ignoring regexp: #{regexp}" if $DEBUG
-      bugs.delete_if {|bug| bug.desc =~ /#{config.ignore_regexp}/}
+      $stderr.puts "Ignoring regexp: #{regexp}" if $DEBUG
+      bugs.delete_if { |bug| bug.desc =~ /#{config.ignore_regexp}/ }
     end
 
     def delete_uninteresting_bugs(bugs)
       # ignoring all bugs but the requested ones
-      bugs.delete_if { |bug| !config.fbugs.include?(bug.bug_number)}
+      bugs.delete_if { |bug| !config.fbugs.include?(bug.bug_number) }
     end
 
     def iterate_fixed_found_version(bts_versions, src_name)
@@ -1132,7 +1182,7 @@ module Factory
       # find out if this version is buggy or not depending on the
       # fixed / found arrays.
 
-      puts " .. checking ver #{ver} against fixed:#{fixed} found:#{found}" if $DEBUG
+      $stderr.puts " .. checking ver #{ver} against fixed:#{fixed} found:#{found}" if $DEBUG
 
       fixed_max_below = nil
       found_max_below = nil
@@ -1156,7 +1206,7 @@ module Factory
         if found_min_above == nil
           # no new version found;
           # which I guess means that the BTS is missing version info
-          puts " ... no found info: I guess I am buggy" if $DEBUG
+          $stderr.puts " ... no found info: I guess I am buggy" if $DEBUG
           val=true
         else
           # found_min_above is not nil, which means I may not be buggy;
@@ -1164,27 +1214,27 @@ module Factory
           # which probably means 'found' info is missing from BTS
           if fixed_min_above == nil ||
               Debian::Dpkg.compare_versions(fixed_min_above, "gt", found_min_above)
-            puts " ... bug found in a later version: I guess I am not buggy" if $DEBUG
+            $stderr.puts " ... bug found in a later version: I guess I am not buggy" if $DEBUG
             val=false
           else
-            puts " ... bug fixed in a later version before it's found again: I guess I am buggy" if $DEBUG
+            $stderr.puts " ... bug fixed in a later version before it's found again: I guess I am buggy" if $DEBUG
             val=true
           end
         end
       else if fixed_max_below == nil
              # the bug was found before (or in) this version, but fixed
              # later (or never): it means I am buggy
-             puts " ... bug found in a prior version, but not yet fixed: I am buggy" if $DEBUG
+             $stderr.puts " ... bug found in a prior version, but not yet fixed: I am buggy" if $DEBUG
              val=true
            else if found_max_below == nil || Debian::Dpkg.compare_versions(fixed_max_below, "gt", found_max_below)
                   # the bug was not found between the latest fixed version
                   # and this version: it means I am not buggy
-                  puts " ... bug not found between the latest fixed version and this version: I am not buggy" if $DEBUG
+                  $stderr.puts " ... bug not found between the latest fixed version and this version: I am not buggy" if $DEBUG
                   val=false
                 else
                   # the bug was indeed found between the latest fixed version
                   # and this version: it means I am buggy
-                  puts " ... bug found between the latest fixed version and this version: I am buggy" if $DEBUG
+                  $stderr.puts " ... bug found between the latest fixed version and this version: I am buggy" if $DEBUG
                   val=true
                 end
            end
@@ -1198,15 +1248,15 @@ module Factory
       # @return false: bug is relevant, true: bug is irrelevant, should be removed.
       val = false
 
-      puts "Start checking: #{bug_number}" if $DEBUG
+      $stderr.puts "Start checking: #{bug_number}" if $DEBUG
 
       # ignore bugs that have no fixed version info, and are closed with a XXXX-done
       if fixed.nil? && bug_stat == "done"
-        puts "bug apparently closed with XXXX-done without version info" if $DEBUG
+        $stderr.puts "bug apparently closed with XXXX-done without version info" if $DEBUG
         val = true
       else if new_ver.nil?
              # no specific version was given for the package, which means that we are interested in all (non archived) bugs
-             puts "no package version given" if $DEBUG
+             $stderr.puts "no package version given" if $DEBUG
              val = false
            else if cur_ver.nil?
                   # no known installed version, which means that we want to check the new version
@@ -1218,7 +1268,7 @@ module Factory
            end
       end
 
-      puts "in conclusion, for bug #{bug_number} comparing fixed:[#{fixed}], found:[#{found}], cur_ver:#{cur_ver} and new_ver:#{new_ver} results in removal:#{val}" if $DEBUG
+      $stderr.puts "in conclusion, for bug #{bug_number} comparing fixed:[#{fixed}], found:[#{found}], cur_ver:#{cur_ver} and new_ver:#{new_ver} results in removal:#{val}" if $DEBUG
       val
     end
 
@@ -1268,13 +1318,13 @@ module Factory
     end
 
     def delete_unwanted_tag_bugs( bugs )
-      puts "checking unwanted bugs: #{bugs}" if $DEBUG
+      $stderr.puts "checking unwanted bugs: #{bugs}" if $DEBUG
       bugs.delete_if { |bug|
         val = false
-        puts "#{bug}" if $DEBUG
+        $stderr.puts "#{bug}" if $DEBUG
         config.tag.each { |tag|
           if bug.tags && bug.tags.include?( tag )
-            puts "#{bug} has {tag}" if $DEBUG
+            $stderr.puts "#{bug} has {tag}" if $DEBUG
           else
             val = true
           end
@@ -1305,13 +1355,13 @@ class ConsoleFrontend
   end
 
   def progress(msg, val)
-    $stderr.print "\r"
-    $stderr.print " " * Unicode.width(@old)
-    $stderr.print "\r"
+    $stdout.print "\r"
+    $stdout.print " " * Unicode.width(@old)
+    $stdout.print "\r"
     @old = "#{msg} #{val}"
-    $stderr.print @old
-    $stderr.flush
-    $stderr.puts "" if Factory.done?(val)
+    $stdout.print @old
+    $stdout.flush
+    $stdout.puts "" if Factory.done?(val)
   end
 
   def puts(msg)
